import glob
import io
import os
import sys
import tarfile
import tempfile
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import click
import requests
from tqdm import tqdm

from tecton._internals import metadata_service
from tecton._internals import model_utils as internal_model_utils
from tecton._internals import type_utils
from tecton.cli import model_utils
from tecton.cli import printer
from tecton.cli.cli_utils import confirm_or_exit
from tecton.cli.cli_utils import display_principal
from tecton.cli.cli_utils import display_table
from tecton.cli.cli_utils import timestamp_to_string
from tecton.cli.command import TectonGroup
from tecton.cli.upload_utils import DEFAULT_MAX_WORKERS_THREADS
from tecton.cli.upload_utils import UploadPart
from tecton.cli.upload_utils import get_upload_parts
from tecton.framework.model_config import ModelConfig
from tecton_core import http
from tecton_core.data_types import data_type_from_proto
from tecton_core.id_helper import IdHelper
from tecton_proto.common import schema__client_pb2 as schema_pb2
from tecton_proto.common.id__client_pb2 import Id
from tecton_proto.modelartifactservice.model_artifact_data__client_pb2 import ModelArtifactInfo
from tecton_proto.modelartifactservice.model_artifact_data__client_pb2 import ModelType
from tecton_proto.modelartifactservice.model_artifact_service__client_pb2 import CompleteModelArtifactUploadRequest
from tecton_proto.modelartifactservice.model_artifact_service__client_pb2 import CreateModelArtifactRequest
from tecton_proto.modelartifactservice.model_artifact_service__client_pb2 import DeleteModelArtifactRequest
from tecton_proto.modelartifactservice.model_artifact_service__client_pb2 import FetchModelArtifactRequest
from tecton_proto.modelartifactservice.model_artifact_service__client_pb2 import GetModelArtifactUploadUrlRequest
from tecton_proto.modelartifactservice.model_artifact_service__client_pb2 import ListModelArtifactsRequest
from tecton_proto.modelartifactservice.model_artifact_service__client_pb2 import UpdateTectonModelRequest
from tecton_proto.modelartifactservice.model_artifact_service__client_pb2 import UploadModelArtifactPartRequest


@click.group("model", cls=TectonGroup)
def model():
    """Manage models."""


@model.command("create")
@click.argument("model_config_file", required=True, type=click.Path(exists=True))
def create(model_config_file):
    """Create a custom model."""
    _create(model_config_file)


@model.command("update")
@click.argument("model_config_file", required=True, type=click.Path(exists=True))
def update(model_config_file):
    """Update a custom model."""
    _update(model_config_file)


def error_and_exit(message: str):
    error_message = f"⛔ {message}"
    printer.safe_print(error_message, file=sys.stderr)
    sys.exit(1)


def try_and_exit(call: Callable[[], Any], message: str) -> Any:
    try:
        return call()
    except Exception as e:
        printer.safe_print(f"{message} : {e}", file=sys.stderr)
        sys.exit(1)


def _resolve_file(model_config_file_path: Path, file: str) -> Path:
    return Path(model_config_file_path / file)


def _create(model_config_file: str):
    """Create a custom model"""
    model_config_file_path = Path(os.path.abspath(click.format_filename(model_config_file)))
    if not os.path.isfile(model_config_file):
        internal_model_utils.error_and_exit(f"{model_config_file} is not a file")

    model_configs = model_utils.get_custom_models(model_config_file_path)
    if len(model_configs) != 1:
        internal_model_utils.error_and_exit(
            f"Exactly one `ModelConfig` should be defined in the python file, provided file has {len(model_configs)}."
        )
    model_config = model_configs[0]
    _create_from_model_config(model_config, model_config_file_path)


def _create_from_model_config(model_config: ModelConfig, model_config_file_path: Path):
    model_config_file_dir = model_config_file_path.parent
    artifact_files = model_config.artifact_files or []

    internal_model_utils.validate(
        model_file=model_config.model_file, archive_root_path=model_config_file_dir, artifact_files=artifact_files
    )
    model_file_path = model_config_file_dir / model_config.model_file
    all_file_paths = [
        *[model_config_file_dir / file for file_str in artifact_files for file in glob.glob(file_str)],
        model_file_path,
    ]

    # Create Model Artifact in DB
    model_artifact_id = create_model_artifact(
        name=model_config.name,
        model_file_path=str(model_file_path.relative_to(model_config_file_dir)),
        model_config_file_path=model_config_file_path.name,
        type=internal_model_utils.model_type_string_to_enum(model_config.model_type),
        description=model_config.description,
        tags=model_config.tags,
        input_schema=type_utils.to_tecton_schema(model_config.input_schema),
        output_schema=type_utils.to_tecton_schema([model_config.output_schema]),
        environments=model_config.environments,
        artifact_files=model_config.artifact_files,
    )
    printer.safe_print(f"Created Model Artifact: {model_config.name}")

    start_request = GetModelArtifactUploadUrlRequest(model_artifact_id=model_artifact_id)
    start_response = try_and_exit(
        lambda: metadata_service.instance().GetModelArtifactUploadUrl(start_request),
        message="Failed to Get Upload URL for Model Artifact",
    )

    printer.safe_print(f"Uploading {len(all_file_paths)} files into model archive.")

    # Upload Model Config
    model_config_upload_url = start_response.model_config_upload_url
    with open(model_config_file_path, "rb") as file_data:
        response = http.session().put(model_config_upload_url, data=file_data)
        if not response.ok:
            msg = f"Upload of model config failed with status {response.status_code} and error {response.text}"
            raise ValueError(msg)

    # Upload Artifact Files
    upload_id = start_response.upload_id
    upload_parts = _upload(
        model_artifact_id=model_artifact_id,
        upload_id=upload_id,
        archive_root_dir=model_config_file_dir,
        files=all_file_paths,
    )

    complete_upload_request = CompleteModelArtifactUploadRequest(
        model_artifact_id=model_artifact_id, upload_id=upload_id, part_etags=upload_parts
    )

    try_and_exit(
        lambda: metadata_service.instance().CompleteModelArtifactUpload(complete_upload_request),
        message="Failed to Upload Model Artifact",
    )

    printer.safe_print(f"✅ Successfully created and uploaded model: {model_config.name}")


def _update(model_config_file: str) -> None:
    model_config_file_path = Path(os.path.abspath(click.format_filename(model_config_file)))
    if not os.path.isfile(model_config_file):
        error_and_exit(f"{model_config_file} is not a file")

    model_configs = model_utils.get_custom_models(model_config_file_path)
    if len(model_configs) != 1:
        error_and_exit(
            f"Exactly one `ModelConfig` should be defined in the python file, provided file has {len(model_configs)}."
        )
    model_config = model_configs[0]
    model_config_file_dir = model_config_file_path.parent
    model_file_path = model_config_file_dir / model_config.model_file

    # Upload Model Config
    model_config_upload_url = update_model_artifact(
        name=model_config.name,
        type=internal_model_utils.model_type_string_to_enum(model_config.model_type),
        description=model_config.description,
        tags=model_config.tags,
        input_schema=type_utils.to_tecton_schema(model_config.input_schema),
        output_schema=type_utils.to_tecton_schema([model_config.output_schema]),
        environments=model_config.environments,
        model_file_path=str(model_file_path.relative_to(model_config_file_dir)),
        model_config_file_path=model_config_file_path.name,
        artifact_files=model_config.artifact_files,
    )
    with open(model_config_file_path, "rb") as file_data:
        response = http.session().put(model_config_upload_url, data=file_data)
        if not response.ok:
            msg = f"Upload of model config failed with status {response.status_code} and error {response.text}"
            raise ValueError(msg)

    printer.safe_print(f"✅ Successfully updated model: {model_config.name}")


def update_model_artifact(
    name: str,
    model_file_path: str,
    model_config_file_path: str,
    type: ModelType,
    input_schema: schema_pb2.Schema,
    output_schema: schema_pb2.Schema,
    environments: List[str],
    tags: Optional[Dict[str, str]] = None,
    description: Optional[str] = None,
    artifact_files: List[str] = [],
) -> str:
    update_model_artifact_request = UpdateTectonModelRequest(
        name=name,
        type=type,
        description=description,
        tags=tags,
        input_schema=input_schema,
        output_schema=output_schema,
        environments=environments,
        model_file_path=model_file_path,
        model_config_file_path=model_config_file_path,
        artifact_files=artifact_files,
    )

    update_model_artifact_response = try_and_exit(
        lambda: metadata_service.instance().UpdateTectonModel(update_model_artifact_request),
        message="Failed to Update Model Artifact",
    )
    model_config_upload_url = update_model_artifact_response.model_config_upload_url
    return model_config_upload_url


def create_model_artifact(
    name: str,
    model_file_path: str,
    model_config_file_path: str,
    type: ModelType,
    input_schema: schema_pb2.Schema,
    output_schema: schema_pb2.Schema,
    environments: List[str],
    tags: Optional[Dict[str, str]] = None,
    description: Optional[str] = None,
    artifact_files: List[str] = [],
) -> Id:
    create_model_artifact_request = CreateModelArtifactRequest(
        name=name,
        model_file_path=model_file_path,
        model_config_file_path=model_config_file_path,
        type=type,
        description=description,
        tags=tags,
        input_schema=input_schema,
        output_schema=output_schema,
        environments=environments,
        artifact_files=artifact_files,
    )

    create_model_artifact_response = try_and_exit(
        lambda: metadata_service.instance().CreateModelArtifact(create_model_artifact_request),
        message="Failed to Create Model Artifact",
    )
    model_artifact_id = create_model_artifact_response.model_artifact_info.id
    return model_artifact_id


def _upload(model_artifact_id: Id, upload_id: str, archive_root_dir: Path, files: List[Path]) -> Dict[int, str]:
    with tempfile.TemporaryDirectory() as tmpdir:
        output_zip_file = Path(tmpdir) / "archive.tar.gz"
        with tarfile.open(output_zip_file, mode="w:gz") as targz:
            for file in files:
                if os.path.islink(file):
                    real_path = os.readlink(file)
                    targz.add(real_path, arcname=Path(file.relative_to(archive_root_dir)), recursive=False)
                else:
                    targz.add(file, arcname=Path(file.relative_to(archive_root_dir)), recursive=False)

        file_size = output_zip_file.stat().st_size
        return dict(
            _upload_file_in_parts(
                file_size=file_size,
                upload_id=upload_id,
                model_artifact_id=model_artifact_id,
                output_zip_file=output_zip_file,
            )
        )


def _upload_file_in_parts(file_size: int, upload_id: str, model_artifact_id: Id, output_zip_file: Path) -> List[Tuple]:
    part_data_list = get_upload_parts(file_size=file_size)
    with ThreadPoolExecutor(DEFAULT_MAX_WORKERS_THREADS) as executor:
        upload_futures = [
            executor.submit(
                _upload_part,
                upload_part=part_data,
                parent_upload_id=upload_id,
                model_artifact_id=model_artifact_id,
                dependency_file_path=output_zip_file,
            )
            for part_data in part_data_list
        ]
        with tqdm(total=len(part_data_list), desc="Upload progress", ncols=100) as pbar:
            for future in as_completed(upload_futures):
                # Increment the tqdm progress bar whenever a future is done
                if future.result():
                    pbar.update(1)

        return [future.result() for future in upload_futures]


def _upload_part(
    upload_part: UploadPart,
    parent_upload_id: str,
    model_artifact_id: Id,
    dependency_file_path: Path,
) -> Tuple[int, str]:
    """Upload a part of a file.

    Args:
        upload_part (UploadPart): The part to upload.
        parent_upload_id (str): The ID of the parent upload.
        model_artifact_id (str): The ID of the Model Artifact.
        dependency_file_path (Path): The path to the file to upload.

    Returns:
        (upload_part, e-tag of that part)
    """
    request = UploadModelArtifactPartRequest(
        model_artifact_id=model_artifact_id, parent_upload_id=parent_upload_id, part_number=upload_part.part_number
    )
    response = metadata_service.instance().UploadModelArtifactPart(request)
    signed_url = response.upload_url

    with open(dependency_file_path, "rb") as fp:
        fp.seek(upload_part.offset)
        file_data = fp.read(upload_part.part_size)
        response = http.session().put(signed_url, data=file_data)
        if response.ok:
            e_tag = response.headers["ETag"]
            return (upload_part.part_number, e_tag)
        else:
            msg = f"Upload failed with status {response.status_code} and error {response.text}"
            raise ValueError(msg)


@click.option("--id", default=None, help="Model Id")
@click.option("-n", "--name", default=None, help="Model Name")
@model.command("list")
def list(id: Optional[str], name: Optional[str]):
    """List custom models."""
    if name and id:
        msg = "Specify either the Model ID or Model Name"
        raise click.ClickException(msg)
    _display_models(_list_models(id, name))


def _list_models(id_string: Optional[str] = None, name: Optional[str] = None):
    id = IdHelper.from_string(id_string) if (id_string) else None
    response = try_and_exit(
        lambda: metadata_service.instance().ListModelArtifacts(ListModelArtifactsRequest(id=id, name=name)),
        message="Failed to fetch models",
    )
    return response.models


def _display_models(models):
    headings = ["Id", "Name", "Created At", "Created By"]
    display_table(
        headings,
        [
            (
                IdHelper.to_string(m.id),
                m.name,
                timestamp_to_string(m.created_at),
                display_principal(m.created_by_principal),
            )
            for m in models
        ],
    )


@click.option("--id", default=None, help="Model Id")
@click.option("-n", "--name", default=None, help="Model Name")
@model.command("describe")
def describe(id: Optional[str], name: Optional[str]):
    """Describe a custom model."""
    if not (bool(name) ^ bool(id)):
        msg = "Specify either the Model ID or Model Name"
        raise click.ClickException(msg)
    models = _list_models(id, name)
    if len(models) != 1:
        printer.safe_print(f"Error: {len(models)} models found.", file=sys.stderr)
        return
    _display_model(models[0])


def _display_model(model: ModelArtifactInfo):
    printer.safe_print(f"{'Name: ': <15}{model.name}")
    printer.safe_print(f"{'ID: ': <15}{IdHelper.to_string(model.id)}")
    if model.description:
        printer.safe_print(f"{'Description: ': <15}{model.description}")
    if model.tags:
        printer.safe_print(f"{'Tags: ': <15}{model.tags}")
    if model.HasField("created_by"):
        printer.safe_print(f"{'Created At: ': <15}{timestamp_to_string(model.created_at)}")
    if model.HasField("created_by_principal"):
        printer.safe_print(f"{'Created By: ': <15}{display_principal(model.created_by_principal)}")
    printer.safe_print(f"{'Environments: ': <15}{model.environments}")
    printer.safe_print()
    printer.safe_print("Input Schema:")
    headings = ["Column Name", "Data Type"]
    display_table(
        headings,
        [(column.name, data_type_from_proto(column.offline_data_type)) for column in model.input_schema.columns],
    )
    printer.safe_print()
    printer.safe_print("Output Schema: ")
    display_table(
        headings,
        [(column.name, data_type_from_proto(column.offline_data_type)) for column in model.output_schema.columns],
    )
    printer.safe_print()


@click.option("--id", default=None, help="Model Id")
@click.option("-n", "--name", default=None, help="Model Name")
@model.command("delete")
def delete(id: Optional[str], name: Optional[str]):
    """Delete custom models."""
    if not (bool(name) ^ bool(id)):
        msg = "Specify either the Model ID or Model Name"
        raise click.ClickException(msg)
    _delete_models(id, name)
    printer.safe_print("✅ Successfully deleted model.")


def _delete_models(id_string: Optional[str] = None, name: Optional[str] = None):
    id = IdHelper.from_string(id_string) if id_string else None
    try_and_exit(
        lambda: metadata_service.instance().DeleteModelArtifact(DeleteModelArtifactRequest(id=id, name=name)),
        message="Failed to delete model",
    )


@click.option("--id", default=None, help="Model Id")
@click.option("-n", "--name", default=None, help="Model Name")
@click.option("-c", "--config-only", default=False, is_flag=True, help="Fetch only Model Config file")
@click.option("--yes", "-y", is_flag=True, default=False, help="Skip all confirmation prompts.")
@model.command("fetch")
def fetch(id: Optional[str], name: Optional[str], config_only: bool, yes: bool):
    """Fetch model artifacts for a custom model."""
    if not (bool(name) ^ bool(id)):
        msg = "Specify either the Model ID or Model Name"
        raise click.ClickException(msg)
    _fetch_model_artifact(id, name, config_only, yes)


def _fetch_model_artifact(id_string: Optional[str], name: Optional[str], config_only: bool, yes: bool):
    try:
        id = IdHelper.from_string(id_string) if id_string else None
        response = metadata_service.instance().FetchModelArtifact(FetchModelArtifactRequest(id=id, name=name))

        try:
            all_downloaded_files = []
            message = ""

            model_config_response = http.session().get(response.model_config_download_url)
            model_config_response.raise_for_status()
            model_config_file_name = os.path.basename(response.model_config_download_url.split("?")[0])
            all_downloaded_files.append(model_config_file_name)

            tar_response = None
            if not config_only:
                tar_response = http.session().get(response.model_artifact_download_url)
                tar_response.raise_for_status()
                with tarfile.open(fileobj=io.BytesIO(tar_response.content), mode="r|gz") as tar:
                    files = tar.getnames()
                    all_downloaded_files += files

            files_in_cur_dir = [file for file in all_downloaded_files if os.path.isfile(file)]
            if len(files_in_cur_dir) > 0 and not yes:
                for f in files_in_cur_dir:
                    printer.safe_print(f)
                confirm_or_exit("This operation may overwrite the files listed above. Ok?")

            # Fetch Model Config
            with open(model_config_file_name, "wb") as file:
                file.write(model_config_response.content)
                message += f"Fetched model config file: {model_config_file_name}"

            # Fetch Model Artifacts
            if not config_only and tar_response:
                with tarfile.open(fileobj=io.BytesIO(tar_response.content), mode="r|gz") as tar:
                    tar.extractall()
                    message += f" and {len(tar.getmembers())} model files"

            printer.safe_print(f"{message} and wrote to current directory")

        except requests.RequestException as e:
            raise SystemExit(e)

    except Exception as e:
        printer.safe_print(f"Failed to fetch model: {e}", file=sys.stderr)
        sys.exit(1)



from types import MappingProxyType
from typing import Callable
from typing import Mapping
from typing import Optional

import attrs
from typeguard import typechecked

from tecton_core import function_deserialization
from tecton_core.specs import tecton_object_spec
from tecton_core.specs import utils
from tecton_proto.args import transformation__client_pb2 as transformation__args_pb2
from tecton_proto.data import transformation__client_pb2 as transformation__data_pb2
from tecton_proto.validation import validator__client_pb2 as validator_pb2


__all__ = ["TransformationSpec"]


@utils.frozen_strict
class TransformationSpec(tecton_object_spec.TectonObjectSpec):
    transformation_mode: transformation__args_pb2.TransformationMode.ValueType
    user_function: Callable = attrs.field(metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True})
    prevent_destroy: bool
    options: Mapping[str, str]

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: transformation__data_pb2.Transformation, include_main_variables_in_scope: bool = False
    ) -> "TransformationSpec":
        user_function = None
        if proto.HasField("user_function"):
            user_function = function_deserialization.from_proto(proto.user_function, include_main_variables_in_scope)
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_data_proto(
                proto.transformation_id, proto.fco_metadata
            ),
            transformation_mode=proto.transformation_mode,
            user_function=user_function,
            prevent_destroy=proto.validation_args.args.prevent_destroy,
            validation_args=validator_pb2.FcoValidationArgs(transformation=proto.validation_args),
            options=MappingProxyType(proto.options),
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: transformation__args_pb2.TransformationArgs, user_function: Optional[Callable]
    ) -> "TransformationSpec":
        # If a function was serialized for this transformation (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.
        if proto.HasField("user_function"):
            user_function = function_deserialization.from_proto(proto.user_function)
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_args_proto(
                proto.transformation_id, proto.info, proto.version
            ),
            transformation_mode=proto.transformation_mode,
            user_function=user_function,
            prevent_destroy=proto.prevent_destroy,
            validation_args=None,
            options=MappingProxyType(proto.options),
        )



from typing import Callable
from typing import Mapping
from typing import Optional

import attrs
from typeguard import typechecked

from tecton_core import function_deserialization
from tecton_core.specs import tecton_object_spec
from tecton_core.specs import utils
from tecton_core.specs.secret_spec import SecretSpec
from tecton_proto.args import resource_provider__client_pb2 as resource_provider__arg_pb2
from tecton_proto.common import framework_version__client_pb2 as framework_version_pb2
from tecton_proto.data import resource_provider__client_pb2 as resource_provider__data_pb2
from tecton_proto.validation import validator__client_pb2 as validator_pb2


@utils.frozen_strict
class ResourceProviderSpec(tecton_object_spec.TectonObjectSpec):
    """Base class for resource provider specs."""

    secrets: Optional[Mapping[str, SecretSpec]]
    function: Callable = attrs.field(metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True})

    @classmethod
    @typechecked
    def from_data_proto(cls, proto: resource_provider__data_pb2.ResourceProvider) -> "ResourceProviderSpec":
        function = function_deserialization.from_proto(proto.function)
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_data_proto(
                proto.resource_provider_id, proto.fco_metadata
            ),
            secrets={key: SecretSpec.from_proto(value) for (key, value) in proto.secrets.items()},
            function=function,
            validation_args=validator_pb2.FcoValidationArgs(resource_provider=proto.validation_args),
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: resource_provider__arg_pb2.ResourceProviderArgs, function: Optional[Callable]
    ) -> "ResourceProviderSpec":
        # If a function was serialized for this resource (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.
        if proto.HasField("function"):
            function = function_deserialization.from_proto(proto.function)
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_args_proto(
                proto.resource_provider_id, proto.info, framework_version_pb2.FrameworkVersion.FWV6
            ),
            secrets={key: SecretSpec.from_proto(value) for (key, value) in proto.secrets.items()},
            function=function,
            validation_args=None,
        )



import contextlib
import logging
from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Union

import attrs
import pyarrow

from tecton_core.query.dialect import Dialect
from tecton_core.query.executor_params import ExecutionContext
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.nodes import StagedTableScanNode
from tecton_core.query.pandas.nodes import ArrowDataNode
from tecton_core.schema import Schema


if TYPE_CHECKING:
    from tecton_core.query.query_tree_executor import QueryTreeExecutor


logger = logging.getLogger(__name__)


@attrs.define
class QueryTreeCompute(ABC):
    """
    Base class for compute (e.g. DWH compute or Python compute) which can be
    used for different stages of executing the query tree.
    """

    @staticmethod
    def for_dialect(
        dialect: Dialect,
        executor: "QueryTreeExecutor",
        qt_root: Optional[NodeRef] = None,
    ) -> "QueryTreeCompute":
        # Conditional imports are used so that optional dependencies such as the Snowflake connector are only imported
        # if they're needed for a query
        if dialect == Dialect.SNOWFLAKE:
            from tecton_core.query.snowflake.compute import SnowflakeCompute
            from tecton_core.query.snowflake.compute import create_snowflake_connection

            if SnowflakeCompute.is_context_initialized():
                return SnowflakeCompute.from_context()
            return SnowflakeCompute.for_connection(create_snowflake_connection(qt_root, executor.secret_resolver))
        if dialect == Dialect.DUCKDB:
            from tecton_core.query.duckdb.compute import DuckDBCompute

            return DuckDBCompute.from_context(
                offline_store_options=executor.offline_store_options_providers, duckdb_config=executor.duckdb_config
            )

        if dialect == Dialect.BIGQUERY:
            from tecton_core.query.bigquery.compute import BigqueryCompute

            return BigqueryCompute()

        if dialect == Dialect.ARROW:
            return ArrowCompute()

        msg = f"Dialect {dialect} is not supported"
        raise ValueError(msg)


@attrs.define
class ComputeMonitor:
    log_progress: Callable[[float], None] = lambda _: _
    set_query: Callable[[str], None] = lambda _: _


@attrs.define
class SQLCompute(QueryTreeCompute, contextlib.AbstractContextManager):
    """
    Base class for compute backed by a SQL engine (e.g. Snowflake and DuckDB).
    """

    @abstractmethod
    def get_dialect(self) -> Dialect:
        pass

    @abstractmethod
    def run_sql(
        self,
        sql_string: str,
        return_dataframe: bool = False,
        expected_output_schema: Optional[Schema] = None,
        monitor: Optional[ComputeMonitor] = None,
        checkpoint_as: Optional[str] = None,
    ) -> Optional[pyarrow.RecordBatchReader]:
        pass

    @abstractmethod
    def register_temp_table(
        self, table_name: str, table_or_reader: Union[pyarrow.Table, pyarrow.RecordBatchReader]
    ) -> None:
        pass

    @abstractmethod
    def unregister_temp_table(self, table_name: str) -> None:
        pass

    def cleanup_temp_tables(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup_temp_tables()


@attrs.define
class ArrowCompute(QueryTreeCompute, contextlib.AbstractContextManager):
    def run(
        self,
        qt_node: NodeRef,
        input_data: Dict[str, pyarrow.RecordBatchReader],
        context: ExecutionContext,
        monitor: Optional[ComputeMonitor] = None,
    ) -> "pyarrow.RecordBatchReader":
        def replace_staging_scan_nodes(tree: NodeRef) -> None:
            if isinstance(tree.node, StagedTableScanNode):
                staged_table_name = tree.node.staging_table_name
                if staged_table_name not in input_data:
                    msg = f"Missing input {staged_table_name}"
                    raise ValueError(msg)

                tree.node = ArrowDataNode(
                    input_reader=input_data[staged_table_name],
                    input_node=None,
                    columns=tree.node.columns,
                    column_name_updater=lambda x: x,
                    output_schema=tree.node.output_schema,
                )
                return

            for i in tree.inputs:
                replace_staging_scan_nodes(i)

        replace_staging_scan_nodes(qt_node)

        return qt_node.to_arrow_reader(context)

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



from urllib.parse import urlparse

import boto3

from tecton_proto.common.id__client_pb2 import Id


class ModelArtifactProvider:
    def get_presigned_url(self, model_artifact_id: Id, s3_path: str) -> str:
        return


class Boto3ModelArtifactProvider(ModelArtifactProvider):
    def get_presigned_url(self, model_artifact_id: Id, s3_path: str) -> str:
        parsed_url = urlparse(s3_path)
        s3_client = boto3.client("s3")
        return s3_client.generate_presigned_url(
            "get_object", Params={"Bucket": parsed_url.netloc, "Key": parsed_url.path.lstrip("/")}
        )


DEFAULT_MODEL_PROVIDER = Boto3ModelArtifactProvider()



from tecton_proto.auditlog import metadata__client_pb2 as _metadata__client_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class Id(_message.Message):
    __slots__ = ["least_significant_bits", "most_significant_bits"]
    LEAST_SIGNIFICANT_BITS_FIELD_NUMBER: _ClassVar[int]
    MOST_SIGNIFICANT_BITS_FIELD_NUMBER: _ClassVar[int]
    least_significant_bits: int
    most_significant_bits: int
    def __init__(self, most_significant_bits: _Optional[int] = ..., least_significant_bits: _Optional[int] = ...) -> None: ...



from tecton_proto.auditlog import metadata__client_pb2 as _metadata__client_pb2
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor
SERVICE_ACCOUNT_CREDENTIALS_TYPE_API_KEY: ServiceAccountCredentialsType
SERVICE_ACCOUNT_CREDENTIALS_TYPE_OAUTH_CLIENT_CREDENTIALS: ServiceAccountCredentialsType
SERVICE_ACCOUNT_CREDENTIALS_TYPE_UNSPECIFIED: ServiceAccountCredentialsType

class MaskedClientSecret(_message.Message):
    __slots__ = ["created_at", "masked_secret", "secret_id", "status", "updated_at"]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    MASKED_SECRET_FIELD_NUMBER: _ClassVar[int]
    SECRET_ID_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    created_at: str
    masked_secret: str
    secret_id: str
    status: str
    updated_at: str
    def __init__(self, secret_id: _Optional[str] = ..., created_at: _Optional[str] = ..., updated_at: _Optional[str] = ..., status: _Optional[str] = ..., masked_secret: _Optional[str] = ...) -> None: ...

class NewClientSecret(_message.Message):
    __slots__ = ["created_at", "secret", "secret_id", "status"]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    SECRET_FIELD_NUMBER: _ClassVar[int]
    SECRET_ID_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    created_at: str
    secret: str
    secret_id: str
    status: str
    def __init__(self, secret_id: _Optional[str] = ..., created_at: _Optional[str] = ..., status: _Optional[str] = ..., secret: _Optional[str] = ...) -> None: ...

class ServiceAccountCredentialsType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



import dataclasses
import logging
import re
import typing
from typing import Iterable
from typing import Optional
from typing import Union

import attrs
import pandas

from tecton_core.duckdb_factory import DuckDBConfig


try:
    import duckdb
except ImportError:
    msg = (
        "Couldn't initialize Rift compute. "
        "To use Rift install all Rift dependencies first by executing `pip install tecton[rift]`."
    )
    raise RuntimeError(msg)
import pyarrow.json
import sqlparse
from duckdb import DuckDBPyConnection

from tecton_core import conf
from tecton_core import duckdb_factory
from tecton_core.errors import TectonValidationError
from tecton_core.offline_store import OfflineStoreOptionsProvider
from tecton_core.query.dialect import Dialect
from tecton_core.query.errors import SQLCompilationError
from tecton_core.query.query_tree_compute import ComputeMonitor
from tecton_core.query.query_tree_compute import SQLCompute
from tecton_core.schema import Schema
from tecton_core.schema_validation import CastError


@dataclasses.dataclass
class _Cause:
    type_name: str
    message: str


_input_error_pattern = re.compile(
    r"Invalid Input Error: arrow_scan: get_next failed\(\): "
    + r"(?:Unknown error|Invalid): (.*)\. Detail: Python exception: (.*)",
    re.DOTALL,
)


def extract_input_error_cause(e: duckdb.InvalidInputException) -> Optional[_Cause]:
    m = _input_error_pattern.match(str(e))
    if m:
        return _Cause(message=m.group(1), type_name=m.group(2))
    else:
        return None


@attrs.define
class DuckDBCompute(SQLCompute):
    session: "DuckDBPyConnection"
    is_debug: bool = attrs.field(init=False)
    created_views: typing.List[str] = attrs.field(init=False)
    offline_store_options: Iterable[OfflineStoreOptionsProvider] = ()

    @staticmethod
    def from_context(
        offline_store_options: Iterable[OfflineStoreOptionsProvider] = (), duckdb_config: Optional[DuckDBConfig] = None
    ) -> "DuckDBCompute":
        return DuckDBCompute(
            session=duckdb_factory.create_connection(duckdb_config), offline_store_options=offline_store_options
        )

    def __attrs_post_init__(self):
        self.is_debug = conf.get_bool("DUCKDB_DEBUG")
        self.created_views = []

    def run_sql(
        self,
        sql_string: str,
        return_dataframe: bool = False,
        expected_output_schema: Optional[Schema] = None,
        monitor: Optional[ComputeMonitor] = None,
        checkpoint_as: Optional[str] = None,
    ) -> Optional[pyarrow.RecordBatchReader]:
        # Notes on case sensitivity:
        # 1. DuckDB is case insensitive when referring to column names, though preserves the
        #    underlying data casing when exporting to e.g. parquet.
        #    See https://duckdb.org/2022/05/04/friendlier-sql.html#case-insensitivity-while-maintaining-case
        #    This means that when using Snowflake for pipeline compute, the view + m13n schema is auto upper-cased
        # 2. When there is a spine provided, the original casing of that spine is used (since DuckDB separately
        #    registers the spine).
        # 3. When exporting values out of DuckDB (to user, or for ODFVs), we coerce the casing to respect the
        #    explicit schema specified. Thus ODFV definitions should reference the casing specified in the dependent
        #    FV's m13n schema.
        sql_string = sqlparse.format(sql_string, reindent=True)
        if self.is_debug:
            logging.warning(f"DUCKDB: run SQL {sql_string}")

        if monitor:
            monitor.set_query(sql_string)

        # Need to use DuckDB cursor (which creates a new connection based on the original connection)
        # to be thread-safe. It avoids a mysterious "unsuccessful or closed pending query result" error too.
        try:
            cursor = self.session.cursor()
            # Although we set timezone globally, DuckDB still needs this cursor-level config to produce
            # correct arrow result. Otherwise, timestamps in arrow table will have a local timezone.
            cursor.sql("SET TimeZone='UTC'")
            duckdb_relation = cursor.sql(sql_string)
            if checkpoint_as:
                duckdb_relation.create(checkpoint_as)
                duckdb_relation = self.session.table(checkpoint_as)

            if return_dataframe:
                res = duckdb_relation.fetch_arrow_reader(batch_size=int(conf.get_or_raise("DUCKDB_BATCH_SIZE")))
            else:
                res = None

            if self.is_debug:
                logging.warning(self.session.sql("FROM duckdb_memory()"))

            return res
        except duckdb.InvalidInputException as e:
            # This means that the iterator we passed into DuckDB failed. If it failed due a TectonValidationError
            # we want to unwrap that to get rid of the noisy DuckDB context which is generally irrelevant to the
            # failure.
            cause = extract_input_error_cause(e)
            if not cause:
                raise
            for error_t in (CastError, TectonValidationError):
                if error_t.__name__ in cause.type_name:
                    raise error_t(cause.message) from None
            raise
        except duckdb.Error as e:
            raise SQLCompilationError(str(e), sql_string) from None

    def get_dialect(self) -> Dialect:
        return Dialect.DUCKDB

    def register_temp_table_from_pandas(self, table_name: str, pandas_df: pandas.DataFrame) -> None:
        self.session.from_df(pandas_df).create_view(table_name)
        self.created_views.append(table_name)

    def register_temp_table(
        self, table_name: str, table_or_reader: Union[pyarrow.Table, pyarrow.RecordBatchReader]
    ) -> None:
        self.session.from_arrow(table_or_reader).create_view(table_name)
        self.created_views.append(table_name)

    def unregister_temp_table(self, table_name: str) -> None:
        self.session.sql(f"DROP TABLE IF EXISTS {table_name}")

    def cleanup_temp_tables(self):
        for view in self.created_views:
            self.session.unregister(view)
        self.created_views = []
